/*************************************************************************\
* Copyright (c) <2016> <Kepco Inc.>
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>

/* include EPICS headers */
#include <cadef.h>

/* Strings describing the connection status of a channel */
	const char *channel_state_str[4] = { "not found","connection lost","connected","closed" };
    	chid SetVoutID,GetVoutID,SetCoutID,GetCoutID,SetOutpID,GetOutpID,MeasVID,MeasCID,SendID,SetParamID,SetValueID;

    	double SetVout,GetVout,SetCout,GetCout,GetVrd,GetCrd,SetOutp,GetOutp; int f,maxf=11;
    	char *Vsval,*Csval; int Osval;  int status,key,runmode; int i=0,imax=0,autorun,scan; int stepdelay,maxloops,runloop;
    	char vstr[5]; char vset[100][10]; int j=0; char startfile[15],runfile[15]; FILE * fptr; char Kepco[10],KInstr[10]; 
	char Kfunc[20][20]; char *func[]={"Send","SetParam","SetValue","SetVolt","GetVolt","SetCurr","GetCurr","SetOutp","GetOutp","MeasVolt","MeasCurr",""};

 
int kbhit(void)
{
  struct termios oldt, newt;
  int ch;
  int oldf;
 
  tcgetattr(STDIN_FILENO, &oldt);
  newt = oldt;
  newt.c_lflag &= ~(ICANON | ECHO);
  tcsetattr(STDIN_FILENO, TCSANOW, &newt);
  oldf = fcntl(STDIN_FILENO, F_GETFL, 0);
  fcntl(STDIN_FILENO, F_SETFL, oldf | O_NONBLOCK);
 
  ch = getchar();
 
  tcsetattr(STDIN_FILENO, TCSANOW, &oldt);
  fcntl(STDIN_FILENO, F_SETFL, oldf);
 
  if(ch != EOF)
  {
    ungetc(ch, stdin);
    return 1;
  }
 
  return 0;
}

void showvals(){
	printf("getting data\n");
    switch (status)
    {
        case ECA_NORMAL:
            printf ("all values received\n");
            break;
        case ECA_TIMEOUT:
            printf ("some values not received\n");
            break;
        default:
            printf ("unexpected error while reading: %s\n",
                ca_message(status));
    }
    
    printf ("Output State	(%s): %g\n", channel_state_str[ca_state(SetOutpID)], GetOutp);
    printf ("Voltage Read	(%s): %g\n", channel_state_str[ca_state(MeasVID)], GetVrd);
    printf ("Current Read	(%s): %g\n", channel_state_str[ca_state(MeasCID)], GetCrd);

}
void SetVolt(char *value){
    printf("setting step %i, '%s' volts\n\n",i,value);
    ca_put(DBR_STRING, SetVoutID, value);
    status = ca_pend_io(1.0); scan=2;
}
void SetCurr(char *value){
    printf("setting %s amps\n\n",value);
    ca_put(DBR_STRING, SetCoutID, value);
    status = ca_pend_io(1.0);
}
void getChanData(){
	if (strlen(Kepco)<2) strcpy(Kepco,"Kepco");
	f=0; while(f<=maxf) { sprintf(Kfunc[f],"%s%s",Kepco,func[f]);  /* printf("func[%i]=%s,",f,Kfunc[f]);// debug */ f++; } f=0; 
	ca_search (Kfunc[f], &SendID);		f++;
	ca_search (Kfunc[f], &SetParamID);	f++;	
	ca_search (Kfunc[f], &SetValueID);	f++;
	ca_search (Kfunc[f], &SetVoutID);	f++;
	ca_search (Kfunc[f], &GetVoutID);	f++;
	ca_search (Kfunc[f], &SetCoutID);	f++;
	ca_search (Kfunc[f], &GetCoutID);	f++;
	ca_search (Kfunc[f], &SetOutpID);	f++;
	ca_search (Kfunc[f], &GetOutpID);	f++;
	ca_search (Kfunc[f], &MeasVID);		f++;
	ca_search (Kfunc[f], &MeasCID); f=0;	
}
void getFile(char *fname){
	if(fname=="") { printf("filename "); scanf("%s",runfile); } else strcpy(runfile,fname); i=0;
	if(strlen(runfile)>1) fptr=fopen(runfile,"r"); else {printf("!");fptr=fopen("Vlist","r");} printf("fptr='%i'\n",fptr);
	fgets(vset[i],10,fptr); if(strlen(vset[i])) vset[i][strlen(vset[i])-1]=0;
	if (vset[0][0]=='K' ) {strcpy(KInstr,vset[0]);printf("KepcoInstr=%s\n",KInstr);}
		else {printf("\nInstrument Prefix "); scanf("%s",KInstr);printf("%i,'%s'",i,vset[i]); i++;} 
	while (fgets(vset[i],10,fptr)) {	if(strlen(vset[i])) vset[i][strlen(vset[i])-1]=0; printf("%i,'%s'",i,vset[i]); i++;}
	imax=i-1; printf("steps=%i\n",imax); fclose(fptr); i=0; 
}
void dokey(){
	char vs[5],cs[5];
  if(key=='0'){
  	ca_put(DBR_STRING, SetOutpID, "0");
	status = ca_pend_io(1.0);
  }
  if(key=='1'){
	ca_put(DBR_STRING, SetOutpID, "1");
	status = ca_pend_io(1.0);
  }
  if(key=='v'){
	printf("new volt ");
	scanf("%s",&vs);
	SetVolt(vs);

    printf("scanning...\n");
    ca_put(DBR_DOUBLE, MeasVID, &GetVrd);
    ca_put(DBR_DOUBLE, MeasCID, &GetCrd);
    status = ca_pend_io(1.0);
    runmode=10;
  }
  if (key=='c'){
	printf("new curr ");
	scanf("%s",&cs);
	SetCurr(cs);

    printf("scanning...\n");
    ca_put(DBR_DOUBLE, MeasVID, &GetVrd);
    ca_put(DBR_DOUBLE, MeasCID, &GetCrd);
    status = ca_pend_io(1.0);
    runmode=10;
  }
  if(key=='~') if(!autorun) {
	if(i>0) { printf("reset step#(%i) (Y/N) ?",i);scanf("%s",&vs);} if(vs[0]=='Y' || vs[0]=='y') i=imax;
	autorun=1; printf("loops(0=infinite) ");scanf("%i",&maxloops);printf("delay(3=~1sec) ");scanf("%i",&stepdelay);
	printf("autorun=%i del=%i loops=%i\n",autorun,stepdelay,maxloops); runloop=0;while(kbhit()) getchar();
  }
  if(key=='!'){ ca_put(DBR_STRING, SendID, "syst:rem"); status = ca_pend_io(1.0); }
  if(key=='r'){
	ca_put(DBR_STRING, SendID, "*RST");
	status = ca_pend_io(1.0);
  }
  if(key=='?'){
    //if(autorun) {rundelay=10;runmode=10;return;}
    printf("scanning...\n");
    ca_put(DBR_DOUBLE, GetOutpID, &GetOutp);
    ca_put(DBR_DOUBLE, MeasVID, &GetVrd);
    ca_put(DBR_DOUBLE, MeasCID, &GetCrd);
    status = ca_pend_io(1.0);
    runmode=10; 
  }
  if(key=='f'){getFile("");i=0;runmode=10;if(!strstr(Kepco,KInstr)) {strcpy(Kepco,KInstr); getChanData();} }
  if(key=='+') printf("\n!\n"); if(key=='m'){if(runmode<10) runmode=10; else runmode=0; printf("runmode=%i",runmode);}
}
int getkey(){
	key=0; 
	if(autorun) {
		printf("running...   loop(%i/%i),step(%i),vset=%s ... press any key to stop\n",runloop,maxloops,i,Vsval);
  		if(stepdelay>=1) usleep(stepdelay*75000); else usleep(50000);
		if(kbhit()) {while(kbhit()) getchar(); autorun=0; runmode=10;} return 0;
	}
	return getchar(); // you may need to adjust these values
}
void waitkey(){
	printf("\n%s:'~'=autorun(%i),'!'=REM,'r'=RST,'0'=Off,'1'=On,'v'=volt,'c'=curr,'?'=read,'+'=Next Step,'m'=stepmode(%i),'f'=Load File 'Q'=Quit,'Enter' = refresh\n",Kepco,autorun,runmode!=10);
	key=getkey(); if(key) dokey();
}
int getConfig(){
	fptr=fopen("KepcoConfig","r"); printf("fptr='%i'\n",fptr); if(!fptr) return 0;
	fgets(startfile,10,fptr); if(strlen(startfile)) startfile[strlen(startfile)-1]=0;
	fclose(fptr); return 1;
}

void ReadVals(){
 key=0; scan=2; // scan value may need to change if reading is out of sink

 while(key!='+' && key!='Q' && runmode>0){

if(scan>0){
    scan--;if(stepdelay>=1) usleep(stepdelay*75000); else usleep(50000);
    printf("scanning...\n");
    ca_put(DBR_DOUBLE, MeasVID, &GetVrd);
    ca_put(DBR_DOUBLE, MeasCID, &GetCrd);
    status = ca_pend_io(1.0);
}
    ca_get(DBR_DOUBLE, GetOutpID, &GetOutp);
    ca_get(DBR_DOUBLE, MeasVID, &GetVrd);
    ca_get(DBR_DOUBLE, MeasCID, &GetCrd);

    printf ("reading ...\n");
    status = ca_pend_io(1.0);
        
    switch (status)
    {
        case ECA_NORMAL:
            printf ("all values received\n");
            break;
        case ECA_TIMEOUT:
            printf ("some values not received\n");
            break;
        default:
            printf ("unexpected error while reading: %s\n",
                ca_message(status));
    }
    
	showvals(); waitkey();

   if(runmode>0 && (runmode<10 || autorun>0)) runmode--;
 }
	
}
int main ()
{
    
     runmode=10; i=0; autorun=0; stepdelay=1; Vsval=vset[0]; Csval="1"; scan=1; maxloops=0;

    /* Step1: initialize channel access and search for all channels. */
	//printf("\nvlist:\n");while(i<imax) {printf("vset[%i]=%s\n",i,vset[i]);i++;}

	if(getConfig()) getFile(startfile); else getFile("!"); strcpy(Kepco,KInstr); getChanData();
	

    ca_task_initialize();

    /* Send all requests in parallel, wait some seconds. */
    printf ("searching ...\n");
    status = ca_pend_io(1.0);
    
    /* Check for errors */
    switch (status)
    {
        case ECA_NORMAL:
            printf ("all channels found\n");
            break;
        case ECA_TIMEOUT:
            printf ("some channels not found yet\n");
            break;
        default:
            printf ("unexpected error while searching: %s\n",
                ca_message(status));
    }

	ca_put(DBR_STRING, SendID, "syst:rem");
	status = ca_pend_io(1.0);
	ca_put(DBR_STRING, SetOutpID, "1");
	SetCurr(Csval);
	status = ca_pend_io(1.0);
	ReadVals();

while(key!='Q'){

	if(i>=imax) {i=0; runloop++; if(maxloops>0 && runloop>maxloops) autorun=0; }
	printf("vset[%i]=%s",i,vset[i]); Vsval=vset[i]; printf("Vsval=%s,vset[%i]=%s\n",Vsval,i,vset[i]);
	if(Vsval[0]=='~' && maxloops==0) {
		if(autorun) autorun=0; else autorun=1;
		if(Vsval[1]) stepdelay=atoi(&Vsval[1]);
		//if(stepdelay<.5) stepdelay=.25;
		printf("autorun=%i del=%i\n",autorun,stepdelay);
	}
	
	else {SetVolt(Vsval);if(stepdelay>=1) usleep(stepdelay*75000); else usleep(50000);}
	
    	ca_put(DBR_STRING, SetOutpID, "1");
    	status = ca_pend_io(1.0);

	if(runmode==0) runmode=5; // monitor a few cycles
    	ReadVals(); if(key=='Q') {printf("bye !\n");goto end;}

	i++;
   }
end:
    /* free all channel access resources */
    ca_task_exit();
    return 0;  
}
